#include <dxconfig.h>
#include "dx/dx.h"
#include "scm.h"
#include "patchlvl.h"
#include <sys/stat.h>

SCM dims2ura(SCM dims,SCM prot,SCM fill);

static Pointer scm_id=NULL;
static time_t lastchanged;
static Object *global_in;
static Object *global_out;

/* convert a simple object to SCM or NULL if not possible */
SCM Object2SCM(Object o) {
  char *string_value;
  int int_value;
  float float_value;
  
  if (DXExtractInteger(o,&int_value)) {
    /* integer */
    return MAKINUM(int_value);
  }
  else if (DXExtractFloat(o,&float_value)) {
    /* float */
    return makflo(float_value);
  }
  else if (DXExtractString(o,&string_value)) {
    /* string */
    return makfrom0str(string_value);
  }
  else {
    /* array */
    int items;
    Type type;
    Category category;
    int rank; 
    int shape;
    SCM array_result=EOL;
    SCM dim_list=EOL;

    if (DXGetArrayInfo((Array)o, &items, &type, &category, &rank, &shape)) {
      /* creer le prototype */
      switch (type) { 
      case TYPE_BYTE: {
	char *data_in;
	int x,y;
	data_in=(char *)DXGetArrayData((Array)o);
	for (y=items-1;y>=0;y--) {
	  SCM x_sh=EOL;
	  if (rank==0) {
	    x_sh=MAKINUM((int)(data_in[y]));
	  } else
	  for (x=shape-1;x>=0;x--) {
	    x_sh=cons(MAKINUM(data_in[y*shape+x]),x_sh);
	  }
	  array_result=cons(x_sh,array_result);
	}
	return array_result;
      } break;
      case TYPE_UBYTE: {
	unsigned char *data_in;
	int x,y;
	data_in=(unsigned char *)DXGetArrayData((Array)o);
	for (y=items-1;y>=0;y--) {
	  SCM x_sh=EOL;
	  if (rank==0) {
	    x_sh=MAKINUM((int)(data_in[y]));
	  } else
	    for (x=shape-1;x>=0;x--) {
	      x_sh=cons(MAKINUM((int)(data_in[y*shape+x])),x_sh);
	    }
	  array_result=cons(x_sh,array_result);
	}
	return array_result;
      } break;
      case TYPE_SHORT: { /* ok */
	short int *data_in;
	int x,y;
	data_in=(short int *)DXGetArrayData((Array)o);
	for (y=items-1;y>=0;y--) {
	  SCM x_sh=EOL;
	  if (rank==0) {
	    x_sh=MAKINUM((int)(data_in[y]));
	  } else
	  for (x=shape-1;x>=0;x--) {
	    x_sh=cons(MAKINUM(data_in[y*shape+x]),x_sh);
	  }
	  array_result=cons(x_sh,array_result);
	}
	return array_result;
      } break;
      case TYPE_USHORT: { /* ok */
	short int *data_in;
	int x,y;
	data_in=(short int *)DXGetArrayData((Array)o);
	for (y=items-1;y>=0;y--) {
	  SCM x_sh=EOL;
	  if (rank==0) {
	    x_sh=MAKINUM((int)(data_in[y]));
	  } else
	  for (x=shape-1;x>=0;x--) {
	    x_sh=cons(MAKINUM(data_in[y*shape+x]),x_sh);
	  }
	  array_result=cons(x_sh,array_result);
	}
	return array_result;
      } break;
      case TYPE_INT: { /* ok */
	int *data_in;
	int x,y;
	data_in=(int *)DXGetArrayData((Array)o);
	for (y=items-1;y>=0;y--) {
	  SCM x_sh=EOL;
	  if (rank==0) {
	    x_sh=MAKINUM((int)(data_in[y]));
	  } else
	  for (x=shape-1;x>=0;x--) {
	    x_sh=cons(MAKINUM(data_in[y*shape+x]),x_sh);
	  }
	  array_result=cons(x_sh,array_result);
	}
	return array_result;
      } break;
      case TYPE_UINT: { /* ok */
	unsigned int *data_in;
	int x,y;
	data_in=(unsigned int *)DXGetArrayData((Array)o);
	for (y=items-1;y>=0;y--) {
	  SCM x_sh=EOL;
	  if (rank==0) {
	    x_sh=MAKINUM((int)(data_in[y]));
	  } else
	  for (x=shape-1;x>=0;x--) {
	    x_sh=cons(MAKINUM(data_in[y*shape+x]),x_sh);
	  }
	  array_result=cons(x_sh,array_result);
	}
	return array_result;
      } break;
      case TYPE_DOUBLE: { /* ok */
	double *data_in;
	int x,y;
	data_in=(double *)DXGetArrayData((Array)o);
	for (y=items-1;y>=0;y--) {
	  SCM x_sh=EOL;
	  if (rank==0) {
	    x_sh=makdbl(data_in[y],0.0);
	  } else
	  for (x=shape-1;x>=0;x--) {
	    x_sh=cons(makdbl(data_in[y*shape+x],0.0),x_sh);
	  }
	  array_result=cons(x_sh,array_result);
	}
	return array_result; 
      } break;
      case TYPE_FLOAT: { /* ok */
	float *data_in;
	int x,y;
	SCM x_sh=EOL;
	data_in=(float *)DXGetArrayData((Array)o);       
	for (y=items-1;y>=0;y--) {
	  x_sh=EOL;
	  if (rank==0) {
	    x_sh=makflo(data_in[y]);
	  } else
	  for (x=shape-1;x>=0;x--) {
	    x_sh=cons(makflo(data_in[y*shape+x]),x_sh);
	  }
	  array_result=cons(x_sh,array_result);
	}
	return array_result;
      } break;
      case TYPE_STRING: {
	char *data;
	char *data_in;
	int base;
	return NULL;
	array_result=dims2ura(dim_list,makfrom0str("a"),EOL);
	data=(char *)VELTS(array_result);
	base=ARRAY_BASE(array_result);
	data_in=(char *)DXGetArrayData((Array)o);
	memcpy(data+base,data_in,items*sizeof(char));
	return array_result;
      } break;
      default:
	return NULL;
      }
    } else    
      return NULL;
  }
}

/*

  in[0]: init file
  in[1]: command to init (setup ...)
  in[2]: command to run  (run .....)
  in[3]: Field   \ 
  in[4]: Field   | converted as an in variables containing all data
  ...            | (((objectname objectvalue) ...) ...)
  in[n]: Field  /

  idem for out

 */

int Convert_In_to_SCM(Object *in, SCM *in_scm) {
  int in_index;
  int field_index;
  /* first process in level */
  SCM in_n_r;

  *in_scm=EOL;

  for (in_index=10;in_index>2;in_index--) {
    Object object_ptr;
    char *object_name;
    char *string_value;
    int integer_value;
    float *float_value;
    
    if (in[in_index]) {

      in_n_r=Object2SCM(in[in_index]);
      
      if (!in_n_r) {
	field_index=0;
	in_n_r=EOL;
	
	while ((object_ptr=DXGetEnumeratedComponentValue(in[in_index],field_index,&object_name))!=0) {
	  SCM scm_comp_value, scm_comp_name;
	  if (!strcmp("positions",object_name)) {
	    int n,n1;
	    int counts[10];
	    float origin[10];
	    float deltas[10];
	    SCM list_values;
	    /* first try if the positions are regular or not */
	    if (DXQueryGridPositions(object_ptr,&n,counts,origin,deltas)) {
	      /* it is regular! 
		 (regular n (counts) (origin) (deltas) ) */

	      scm_comp_value=EOL;
	      /* deltas list */
	      list_values=EOL;
	      n1=n;
	      while (n1>0) {
		list_values=cons(makflo(deltas[n1-1]),list_values);
		n1--;
	      }
	      scm_comp_value=cons(list_values,scm_comp_value);
	      /* origin list */
	      list_values=EOL;
	      n1=n;
	      while (n1>0) {
		list_values=cons(makflo(origin[n1-1]),list_values);
		n1--;
	      }
	      scm_comp_value=cons(list_values,scm_comp_value);
	      /* count list */
	      list_values=EOL;
	      n1=n;
	      while (n1>0) {
		list_values=cons(MAKINUM(counts[n1-1]),list_values);
		n1--;
	      }
	      scm_comp_value=cons(list_values,scm_comp_value);
	      /* n */
	      scm_comp_value=cons(MAKINUM(n),scm_comp_value);
	      scm_comp_name=makfrom0str("regular");
	      scm_comp_value=cons(scm_comp_name,scm_comp_value);
	      scm_comp_name=makfrom0str(object_name);
	      scm_comp_value=cons(scm_comp_name,scm_comp_value);	   
	    } else {
	      /* it is not regular. Just build the list 
	         (irregular ( ) ( ) ... ) */
	      scm_comp_value=Object2SCM(object_ptr);
	      
	      if (!scm_comp_value) {
		scm_comp_value=EOL;
	      }
	
	      scm_comp_value=cons(scm_comp_value,EOL);
	      scm_comp_name=makfrom0str("irregular");
	      scm_comp_value=cons(scm_comp_name,scm_comp_value);
	      scm_comp_name=makfrom0str(object_name);
	      scm_comp_value=cons(scm_comp_name,scm_comp_value);	   
	    }

	  } else if (!strcmp("connections",object_name)) {
	    int n,n1;
	    int counts[10];
	    SCM list_values;
	    /* first try if the positions are regular or not */
	    if (DXQueryGridConnections(object_ptr,&n,counts)) {
	      /* it is regular! 
		 (regular n (counts)) */

	      scm_comp_value=EOL;
	      /* count list */
	      list_values=EOL;
	      n1=n;
	      while (n1>0) {
		list_values=cons(MAKINUM(counts[n1-1]),list_values);
		n1--;
	      }
	      scm_comp_value=cons(list_values,scm_comp_value);
	      /* n */
	      scm_comp_value=cons(MAKINUM(n),scm_comp_value);
	      scm_comp_name=makfrom0str("regular");
	      scm_comp_value=cons(scm_comp_name,scm_comp_value);
	      scm_comp_name=makfrom0str(object_name);
	      scm_comp_value=cons(scm_comp_name,scm_comp_value);	   
	    } else {
	      /* it is not regular. Just build the list 
	         (irregular ( ) ( ) ... ) */
	      scm_comp_value=Object2SCM(object_ptr);
	      
	      if (!scm_comp_value) {
		scm_comp_value=EOL;
	      }
	
	      scm_comp_value=cons(scm_comp_value,EOL);
	      scm_comp_name=makfrom0str("irregular");
	      scm_comp_value=cons(scm_comp_name,scm_comp_value);
	      scm_comp_name=makfrom0str(object_name);
	      scm_comp_value=cons(scm_comp_name,scm_comp_value);	   
	    }
	    } else { /* general case */
	    /*	  DXMessage(" Component %s ",object_name);*/
	    scm_comp_value=Object2SCM(object_ptr);
	    
	    if (!scm_comp_value) {
	      scm_comp_value=EOL;
	    }
	
	    scm_comp_value=cons(scm_comp_value,EOL);
	    scm_comp_name=makfrom0str(object_name);
	    scm_comp_value=cons(scm_comp_name,scm_comp_value);
	    
	  }

	  in_n_r=cons(scm_comp_value,in_n_r);
	  field_index++;
	}
      }
      *in_scm=cons(in_n_r,*in_scm);
    }
  }
} /* Convert_In_to_SCM */

int scm_type(SCM obj) {
  /* type  return
     integer    0
     string     1
     float      2
     array      3
     list       4 
     double     5 
     ?        6 
  */

  if (INUMP(obj)) return 0;
  if (STRINGP(obj) || SYMBOLP(obj)) return 1;
  if (SINGP(obj)) return 2;
  if (ARRAYP(obj)) return 3;
  if (ilength(CAR(obj))>=0) return 4;
  if (NIMP(obj) && INEXP(obj)) return 5;
  return 6;

}

int Convert_SCM_to_Out(Object *out, SCM *out_scm) {
  SCM empty_list;
  int n_out=0;

  empty_list=EOL;

  while (out_scm!=empty_list) {
    switch (scm_type(CAR(out_scm))) {     /* analyse le CAR */
    case 0 : out[n_out]=DXMakeInteger(INUM(CAR(out_scm))); /* integer */
      break;
    case 1 : { 	/* float */
      float f;
      f=FLO(CAR(out_scm));
      out[n_out]=DXMakeFloat(f);
    }
    break;
    case 2 : {      /* string */
      char *tmp_str;
      tmp_str=malloc(LENGTH(CAR(out_scm))+1);
      memcpy(tmp_str,CHARS(CAR(out_scm)),LENGTH(CAR(out_scm)));
      tmp_str[LENGTH(CAR(out_scm))]=0;
      out[n_out]=DXMakeString(tmp_str);
      free(tmp_str);
    }
    break;
    case 3: 
      DXMessage(" array? ");
      out[n_out]=NULL; /* array */
      break;
    case 4:  { /* liste - if the car of the list is a list than make a list.
		  ifnot, make an array */
      if ((CAR(out_scm)!=empty_list) && (scm_type(CAR(CAR(CAR(out_scm))))!=1)) {
	char *tmp_str;
	SCM scm_data;
	Array dxar;
	int nitems;
	int shape;
	int i,j;
	scm_data=CAR(out_scm);
	nitems=ilength(scm_data);
	shape=ilength(CAR(scm_data));
	switch (scm_type(CAR(CAR(scm_data)))) {
	case 0: { /* integer */
	  int * ptr;
	  dxar=DXNewArray(TYPE_INT, CATEGORY_REAL, 1, shape);
	  if (!DXAddArrayData(dxar,0,nitems,NULL)) DXMessage("Erreur 1");
	  ptr=DXGetArrayData(dxar);
	  for (i=0; i<nitems; i++) {
	    SCM tmp_vect;
	    tmp_vect=CAR(scm_data);
	    for (j=0; j<shape; j++) {
	      ptr[j+i*shape]=INUM(CAR(tmp_vect));
	      tmp_vect=CDR(tmp_vect);
	    }
	    scm_data=CDR(scm_data);
	  }
	}
	case 1: /* string */
	  DXMessage(" String? ");
	  dxar=NULL;
	  break;
	case 2: { /* float */
	  float * ptr;
	  dxar=DXNewArray(TYPE_FLOAT, CATEGORY_REAL, 1, shape);
	  if (!DXAddArrayData(dxar,0,nitems,NULL)) DXMessage("Erreur 2");
	  ptr=DXGetArrayData(dxar);
	  for (i=0; i<nitems; i++) {
	    SCM tmp_vect;
	    tmp_vect=CAR(scm_data);
	    for (j=0; j<shape; j++) {
	      float v;
	      v=FLO(CAR(tmp_vect));
	      ptr[j+i*shape]=v;
	      tmp_vect=CDR(tmp_vect);
	    }
	    scm_data=CDR(scm_data);
	  }
	} break;
	case 3: /* array */
	  DXMessage(" array2? ");
	  dxar=NULL;
	  break;
	case 4: /* list */
	  DXMessage(" list?");
	  dxar=NULL;
	  break;
	case 5: { /* double */
	  float * ptr;
	  dxar=DXNewArray(TYPE_FLOAT, CATEGORY_REAL, 1, shape);
	  if (!DXAddArrayData(dxar,0,nitems,NULL)) DXMessage("Erreur 2");
	  ptr=DXGetArrayData(dxar);
	  for (i=0; i<nitems; i++) {
	    SCM tmp_vect;
	    tmp_vect=CAR(scm_data);
	    for (j=0; j<shape; j++) {
	      float v;
	      v=(float)REALPART(CAR(tmp_vect));
	      ptr[j+i*shape]=v;
	      tmp_vect=CDR(tmp_vect);
	    }
	    scm_data=CDR(scm_data);
	  }
	}
	}
	out[n_out]=dxar;
      } else {
	Field newfield;
	SCM scm_field;
	scm_field=CAR(out_scm);
	newfield=DXNewField();
	while (scm_field!=empty_list) {
	  char *tmp_str;
	  SCM scm_component;
	  SCM scm_data;
	  Array dxar;
	  int nitems;
	  int shape;
	  int i,j;
	  scm_component=CAR(scm_field);
	  tmp_str=malloc(LENGTH(CAR(scm_component))+1);
	  memcpy(tmp_str,CHARS(CAR(scm_component)),LENGTH(CAR(scm_component)));
	  tmp_str[LENGTH(CAR(scm_component))]=0;
	  scm_data=CDR(scm_component);
	  scm_data=CAR(scm_data);
	  nitems=ilength(scm_data);
	  shape=ilength(CAR(scm_data));
	  switch (scm_type(CAR(CAR(scm_data)))) {
	  case 0: { /* integer */
	    int * ptr;
	    dxar=DXNewArray(TYPE_INT, CATEGORY_REAL, 1, shape);
	    if (!DXAddArrayData(dxar,0,nitems,NULL)) DXMessage("Erreur 1");
	    ptr=DXGetArrayData(dxar);
	    for (i=0; i<nitems; i++) {
	      SCM tmp_vect;
	      tmp_vect=CAR(scm_data);
	      for (j=0; j<shape; j++) {
		ptr[j+i*shape]=INUM(CAR(tmp_vect));
		tmp_vect=CDR(tmp_vect);
	      }
	      scm_data=CDR(scm_data);
	    }
	  } break;
	  case 1: /* string */
	    DXMessage(" string2?");
	    dxar=NULL;
	    break;
	  case 2: { /* float */
	    float * ptr;
	    dxar=DXNewArray(TYPE_FLOAT, CATEGORY_REAL, 1, shape);
	    if (!DXAddArrayData(dxar,0,nitems,NULL)) DXMessage("Erreur 2");
	    ptr=DXGetArrayData(dxar);
	    for (i=0; i<nitems; i++) {
	      SCM tmp_vect;
	      tmp_vect=CAR(scm_data);
	      for (j=0; j<shape; j++) {
		float v;
		v=FLO(CAR(tmp_vect));
		ptr[j+i*shape]=v;
		tmp_vect=CDR(tmp_vect);
	      }
	      scm_data=CDR(scm_data);
	    }
	  } break;
	  case 3: /* array */
	    DXMessage(" array3?");
	    dxar=NULL;
	    break;
	  case 4: /* list */
	    DXMessage(" list2?");
	    dxar=NULL;
	    break;
	  case 5: { /* double */
	    float * ptr;
	    dxar=DXNewArray(TYPE_FLOAT, CATEGORY_REAL, 1, shape);
	    if (!DXAddArrayData(dxar,0,nitems,NULL)) DXMessage("Erreur 2");
	    ptr=DXGetArrayData(dxar);
	    for (i=0; i<nitems; i++) {
	      SCM tmp_vect;
	      tmp_vect=CAR(scm_data);
	      for (j=0; j<shape; j++) {
		float v;
		v=(float)REALPART(CAR(tmp_vect));
		ptr[j+i*shape]=v;
		tmp_vect=CDR(tmp_vect);
	      }
	      scm_data=CDR(scm_data);
	    }
	  } break;
	  }
	  if (!DXSetComponentValue((Field)newfield,tmp_str,(Object)dxar)) DXMessage("Erreur 3");
	  scm_component=CDR(scm_component);
	  scm_component=CDR(scm_component);
	  while (scm_component!=empty_list) {
	    char *param1, *param2;
	    param1=malloc(LENGTH(CAR(scm_component))+1);
	    memcpy(param1,CHARS(CAR(scm_component)),LENGTH(CAR(scm_component)));
	    param1[LENGTH(CAR(scm_component))]=0;
	    scm_component=CDR(scm_component);
	    param2=malloc(LENGTH(CAR(scm_component))+1);
	    memcpy(param2,CHARS(CAR(scm_component)),LENGTH(CAR(scm_component)));
	    param2[LENGTH(CAR(scm_component))]=0;
	    scm_component=CDR(scm_component);
	    DXSetStringAttribute(dxar,param1,param2);
	  }
	  free(tmp_str);
	  scm_field=CDR(scm_field);
	}
	if (!DXEndField(newfield)) DXMessage("Erreur 4");
	out[n_out]=newfield;
      }
      break;
    }
    case 6 : { 	/* float */
      float f;
      f=(double)REALPART(CAR(out_scm));
      out[n_out]=DXMakeFloat(f);
    }
    default:
      out[n_out]=NULL;
    }
    /* passe au CDR */
    out_scm=CDR(out_scm);
    n_out++;
  }


} /* Convert_SCM_to_Out */

SCM user_main() {
  return MAKINUM(1);
}

SCM call_again() {
  DXReadyToRun(scm_id);
  return UNSPECIFIED;
}

SCM get_dx_input() {
  SCM SCM_IN;
  /* set input in scm */
  Convert_In_to_SCM(global_in,&SCM_IN);
  sysintern("*in*",SCM_IN);
}

SCM set_dx_output() {
  SCM SCM_OUT;
  /* get output from scm */
  SCM_OUT=scm_evstr("*out*");
  Convert_SCM_to_Out(global_out,SCM_OUT); 
}

SCM scmDXMessage(SCM str) {
  char *tmp_str;
  tmp_str=malloc(LENGTH(str)+1);
  memcpy(tmp_str,CHARS(str),LENGTH(str));
  tmp_str[LENGTH(str)]=0;
  DXMessage(tmp_str);
  free(tmp_str); 
  return EOL;
}

Error m_scm(Object *in, Object *out) {
  SCM thedefine;
  static char *initfile, fileprog[200], *firstrun;
  struct stat Buffer;
  time_t changed; 

  global_in=in;
  global_out=out;

  if (!scm_id) { /* first time */
    int buf0;

    scm_id=DXGetModuleId();
    if (!scm_id) {
      out[0]=NULL;
      return ERROR;
    }
    /* scheme initialisation */
    
    init_sbrk(); 

    /*    buf0=init_buf0(NULL);*/
    buf0=0;

    init_scm(0,buf0,0L); /* 0L est la taille de la pile */
    /* scm_init_from_argv(0,NULL,0L,0,0); */

    /* load initialisation file */
    if (!in[0]) {
      DXSetError(ERROR_MISSING_DATA,"missing scm initialisation file");
      return ERROR;
    }

    if (!DXExtractString(in[0],&initfile)) {
      DXSetError(ERROR_BAD_PARAMETER,"initfile must be a string");
      return ERROR;
    }

    /* scm_top_level(initfile, user_main); */
    exitval = MAKINUM(EXIT_SUCCESS);
    errjmp_bad = (char *)0;
    /*errjmp_recursive = 0;*/
    lflush(sys_errp);
    errno = 0;
    SIG_deferred = 0;
    deferred_proc = 0;
    ints_disabled = 0;
    
    scm_ldfile(initfile);

    make_subr("call-again",tc7_subr_0,call_again);
    make_subr("dxmessage",tc7_subr_1,scmDXMessage);
    make_subr("get-dx-input",tc7_subr_0,get_dx_input);
    make_subr("set-dx-output",tc7_subr_0,set_dx_output);


    /* load initialisation file */
    if (in[1]) {
      char *flname;
      if (!DXExtractString(in[1],&flname)) {
	DXSetError(ERROR_BAD_PARAMETER,"initfile must be a string");
	return ERROR;
      }
      strcpy(fileprog,flname);
      scm_ldfile(fileprog);
      stat(fileprog, &Buffer);
      lastchanged=Buffer.st_mtime;
    }


  }

  if (in[1]) {
    char *flname;
    if (!DXExtractString(in[1],&flname)) {
      DXSetError(ERROR_BAD_PARAMETER,"initfile must be a string");
      return ERROR;
    }
    stat(flname, &Buffer);
    changed=Buffer.st_mtime;
    if ((strcmp(flname,fileprog)) || (lastchanged!=changed)) {
      DXMessage(" Init file modified, reloading %s %s ",fileprog,flname);
      lastchanged=changed;
      strcpy(fileprog,flname);
      scm_ldfile(fileprog);
      DXMessage(" Init OK ");
    }
  }

  /* run the program */
  /* get the command from in[1] */

  if (!in[2]) {
    DXSetError(ERROR_MISSING_DATA,"missing scm command");
    return ERROR;
  }

  if (!DXExtractString(in[2],&firstrun)) {
    DXSetError(ERROR_BAD_PARAMETER,"scm command must be a string");
    return ERROR;
  }

  if (firstrun[0]=='(')
    scm_evstr(firstrun);
  else
    scm_ldfile(firstrun);


  return OK;
} /* m_scm */

/* init_user_scm() is called by the scheme procedure
   SCM_INIT_EXTENSIONS in "Init5xx.scm" */
void init_user_scm()
{
  /* Put calls to your C initialization routines here. */
}

