dnl  DX_INSTALL_PATH
dnl  Tries to find the location where dx is installed if it
dnl  can not then it defaults to /usr/local/dx
dnl  --------------------------------------------------------
AC_DEFUN(DX_INSTALL_PATH,
[
AC_CACHE_CHECK([for dx install path], ac_cv_dx_install_path,
[
AC_MSG_RESULT(locating)
DX_DEFAULT_INST=/usr/local/dx
AC_CHECK_PROGS( DX, dx )

DXINST=""
if test -n "$DX" ; then
  AC_MSG_CHECKING([for path via "dx -whereami"])
  DXINST=`$DX -whereami | grep "installed in" | sed -e "s/installed in //" -e "s'^\(.*\)\/$'\1'"`
  if test -z "$DXINST" ; then
        AC_MSG_RESULT([warning: old version of dx script in path])
  elif test "x$ARCH" = "xintelnt" ; then
	DXINST=`cygpath -w -s "$DXINST"`
  fi
fi

if test -z "$DXINST" ; then
  AC_MSG_CHECKING([for /usr/local/bin/dx])
  if test -x "/usr/local/bin/dx" ; then
     DXINST=`/usr/local/bin/dx -whereami | grep "installed in" | sed -e "s/installed in //"`
  fi

  if test -z "$DXINST" ; then
        DXINST=$DX_DEFAULT_INST
        AC_MSG_RESULT([defaulting to $DXINST])
        AC_MSG_WARN([Missing dx script--please install OpenDX first.])
  elif test "x$ARCH" = "xintelnt" ; then
	DXINST=`cygpath -w -s "$DXINST"`
  fi
fi
ac_cv_dx_install_path=$DXINST
])
DXINST=$ac_cv_dx_install_path
])


dnl  DX_GET_PREFIX
dnl  Sets the prefix where to install the samples. Is overridden
dnl  by the --prefix configure line.
dnl  --------------------------------------------------------
AC_DEFUN(DX_GET_PREFIX,
[
DX_DEFAULT_INST=/usr/local/dx
if test "x${prefix}" = "xNONE" ; then
  if test -n "$DXINST" ; then
	if test "x$ARCH" = "xintelnt" ; then
        	prefix=`cygpath -u $DXINST`
	else
        	prefix=`echo $DXINST`
	fi
  else
        prefix=$DX_DEFAULT_INST
  fi
else
  AC_MSG_RESULT(using --prefix value $prefix)
fi
])


dnl
dnl  Check whether using GNU C
dnl  -------------------------------------------------------------
AC_DEFUN(DX_PROG_CC_GNU,
[AC_CACHE_CHECK(whether we are using GNU C, ac_cv_prog_gcc,
[dnl The semicolon is to pacify NeXT's syntax-checking cpp.
cat > conftest.c <<EOF
#ifdef __GNUC__
  yes;
#endif
EOF
if AC_TRY_COMMAND(${CC-cc} -E conftest.c) | egrep yes > foo 2>&1; then
  ac_cv_prog_gcc=yes
else
  ac_cv_prog_gcc=no
fi
rm foo
])])



dnl
dnl  Determine the architecture that this is being run on.
dnl  -------------------------------------------------------------
AC_DEFUN(DX_ARCHITECTURE,
[
    AC_MSG_CHECKING(architecture type)
    if test "$ARCH" = "" ; then
        unameS=`uname -s`
        unameM="`uname -m`"
        ARCH=unknown
        if test $unameS = "FreeBSD" ; then
            ARCH=freebsd
        elif test `echo $unameS | tr A-Z a-z | sed "s/^.*cygwin.*$/yes/"` = "yes" ; then
            ARCH=cygwin
        elif test $unameS = "Linux" ; then
            ARCH=linux
        elif test $unameS = "IRIX" || test $unameS = "IRIX64" ; then
            ARCH=sgi
        elif test $unameS = "AIX" ; then
            ARCH=ibm6000
        elif test "$unameM" = "alpha" ; then
            ARCH=alphax
        elif test $unameS = "HP-UX" ; then
            ARCH=hp700
        elif test $unameS = "SunOS" ; then
            ARCH=solaris
        elif test $unameS = "Darwin" ; then
            ARCH=macos
        fi

    fi
    AC_MSG_RESULT($ARCH)
    AC_SUBST(ARCH)
    AC_DEFINE_UNQUOTED($ARCH)
])


dnl
dnl  With the architecture determine flags needed to compile.
dnl  -------------------------------------------------------------
AC_DEFUN(DX_LOADABLE_MODULE,
[
    if test $ARCH = "intelnt" ; then
	CFLAGS="$CFLAGS -I\"${DXINST}\include\" -module" 
	LIBS="$LIBS -L\"${DXINST}\lib_$ARCH\" dxexec.lib"
    else
	DXINST=$ac_cv_dx_install_path
	LIBS="$LIBS -L$DXINST/lib_$ARCH"
	CFLAGS="$CFLAGS -I$DXINST/include"
	if test $ac_cv_prog_gcc = "yes" ; then
		MDXLDFLAG="-Wl,"
	else
		MDXLDFLAG=""
	fi
	if test $ARCH = "cygwin" ; then
		CFLAGS="$CFLAGS -DDXD_WIN -DWIN32 -D_X86_ -DNOMENUS -nologo -w -0d -LD -G5"
		LDFLAGS="$LDFLAGS -DLL -SUBSYSTEM:console -INCREMENTAL:no -MACHINE:I386 -NOLOGO"
		SYSLIBS="$SYSLIBS kernel32.lib user32.lib gdi32.lib winspool.lib comdlg32.lib advapi32.lib shell32.lib uuid.lib wsock32.lib"
		OLIBS="$OLIBS $DXINST/lib/DXExport.lib"
	fi
	if test $ARCH = "hp700" ; then
		CFLAGS="$CFLAGS -Dhp700 -Aa +z"
		LDFLAGS="$LDFLAGS ${MDXLDFLAG}-q ${MDXLDFLAG}-b ${MDXLDFLAG}-E ${MDXLDFLAG}-eDXEntry"
		SYSLIBS="$SYSLIBS -ldld"
	fi
	if test $ARCH = "ibm6000" ; then
		CFLAGS="$CFLAGS -Dibm6000"
		LDFLAGS="$LDFLAGS ${MDXLDFLAG}-eDXEntry ${MDXLDFLAG}-bI:$DXINST/lib/dxexec.exp"
	fi
	if test $ARCH = "sgi" ; then
		CFLAGS="$CFLAGS -Dsgi"
		LDFLAGS="$LDFLAGS -call_shared -e DXEntry -U -exported_symbol DXEntry"
	fi
	if test $ARCH = "solaris" ; then
		CFLAGS="$CFLAGS -Dsolaris"
		LDFLAGS="$LDFLAGS ${MDXLDFLAG}-G ${MDXLDFLAG}-eDXEntry"
	fi
	if test $ARCH = "alphax" ; then
		CFLAGS="$CFLAGS -Dalphax"
		LDFLAGS="$LDFLAGS -shared -all -e DXEntry -expect_unresolved main -expect_unresolved DX*"
	fi
	if test $ARCH = "linux" ; then
		CFLAGS="$CFLAGS -D_GNU_SOURCE -Dlinux"
		LDFLAGS="$LDFLAGS --shared -eDXEntry"
	fi
	if test $ARCH = "freebsd" ; then
		CFLAGS="$CFLAGS -D_GNU_SOURCE -Dfreebsd"
		LDFLAGS="$LDFLAGS --shared -eDXEntry"
	fi
	if test $ARCH = "macos" ; then
		CFLAGS="$CFLAGS -D_GNU_SOURCE -Dmacos -dynamic"
		LDFLAGS="$LDFLAGS -bundle -undefined suppress -flat_namespace -Wl,-iDXEntry:_DXEntry -e DXEntry"
	fi
    fi
])
