;; V 1.0
;; This program simulate particule stracking in a porous structure.
;; A 3D structure representing a porous media has been analysed by
;; DXSoil module (http://www.bondy.ird.fr/~delerue/dxsoil/demo/).
;; A pore network has been extracted.
;; In this simulation, 2 particules are flowing through the pore
;; space, going from the center of a pore to the center of the next
;; one, follow a gradient of pression.

(require 'random)

;; necessary */
(define *in* '())

;; necessary */
(define *out* '())

;; the turn number in the simulation, usefull if you want to
;; save images 
(define turn 0)

;; some necessary function to find data in the input 
(define (listeq? a b)
  (cond ((and (null? a) (null? b)) #t)
	((and (null? a) (not (null? b))) #f)
	((and (not (null? a)) (null? b)) #f)
	((eq? (car a) (car b)) (listeq? (cdr a) (cdr b)))
	(else #f))
  )

(define (stringeq? a b)
  (listeq? (string->list a) (string->list b))
  )

(define (look-string lst str)
  (if (null? lst) '()
      (if (stringeq? (caar lst) str)
	  (cadar lst)
	  (look-string (cdr lst) str))))

;; dxmessage can only display a string. Here is a funtion to 
;; display a list. Not that dxmessage will bug if the string is 
;; too long. The problem come from OpenDX. 
(define (dxdisplay-sub l)
  (cond ((list? l) (string-append "(" (apply string-append (map dxdisplay-sub l)) ")"))
	((number? l) (string-append (number->string l) " "))
	((string? l) (string-append "\"" l "\" "))
	((boolean? l) (if (eq? l #t) "#t " "#f "))
	((char? l)  l)
	((symbol? l) (string-append "'" (symbol->string l) " "))
	(#t "unkwon type")))

(define (dxdisplay l)
  (dxmessage (dxdisplay-sub l)))

;; select an entrance pore for a particlue 
(define (select-input-pore in)
  ;; return the number of an input pore
  (letrec ((selectin (lambda (a b) 
		      (cond ((null? b) '())
			    ((eq? (caar b) 1) (cons (caar a) (selectin (cdr a) (cdr b))))
			    (else (selectin (cdr a) (cdr b))))))
	   (pore_id (look-string (car in) "pore_id"))
	   (pore_in (look-string (car in) "pore_in"))
	   )
    (let ((pore_list (selectin pore_id pore_in)))
      (list-ref pore_list (random (length pore_list)))
      )
    )
  )

(define (get-pore-pressure n in)
;; n can be a list or the umber of a pore.
;; it return the pressure or list of pressure
  (let ((pore_id (look-string (car in) "pore_id"))
	(pore_pres (look-string (car in) "pressure")))
    (letrec ((get-pressure (lambda (n l1 l2)
			     (cond ((null? l1) '()) ;; no pore number n
				   ((eq? (caar l1) n) (caar l2))
				   (else (get-pressure n (cdr l1) (cdr l2)))))))
      (get-pressure n pore_id pore_pres)
      )
    )
  )
			 
(define (select-next-pore actual in)
  ;; return the number of the next pore 
  ;; (choosen randomly from neighboors of actual)
  ;; id pore in pore_conn must be increased by 1
  (letrec ((select-next (lambda (l l1 p)
			  (cond ((null? l) '())
				((and (eq? (caar l) (- actual 1)) (> (get-pore-pressure (+ (cadar l) 1) in) p)) 
				 (cons (list (cadar l) (car l1)) (select-next (cdr l) (cdr l1) p)))
				((and (eq? (cadar l) (- actual 1)) (> (get-pore-pressure (+ (caar l) 1) in) p)) 
				 (cons (list (caar l) (car l1)) (select-next (cdr l) (cdr l1) p)))
				(else (select-next (cdr l) (cdr l1) p)))))
	   (pore_conn (look-string (car in) "pore_conn"))
;	   (pore_pres (look-string (car in) "pressure"))
	   (pore_int (look-string (car in) "pore_int"))
	   )
    (let ((pore_list (select-next pore_conn pore_int (get-pore-pressure actual in))))
      (if (null? pore_list)
	  '()
	  (apply (lambda (x y) (list (+ x 1) y)) (list-ref pore_list (random (length pore_list))))
	  )
      )
    )
  )

(define (get-pore-position n in)
  ;; return the po
  (letrec ((select-position (lambda (a b) 
			      (cond ((null? a) '())
				    ((eq? (caar a) n) (car b))
				    (else (select-position (cdr a) (cdr b))))))
	   (pore_id (look-string (car in) "pore_id"))
	   (pore_center (look-string (car in) "pore_center"))
	   )
    (select-position pore_id pore_center)))

(define (create-particule )
  ;; Create a particule object.
  (let* ((scale-m  3)
	 (actual-pore (select-input-pore *in*))
	 (next-dbl (select-next-pore actual-pore *in*))
	 (next-pore (car next-dbl))
	 (pore-position (get-pore-position actual-pore *in*))
	 (next-position (list (cadr next-dbl) (get-pore-position next-pore *in*)))
	 (percent 0)
	 (plength (* scale-m (sqrt (apply + (map (lambda (p1 p2) (* (- p2 p1) (- p2 p1))) pore-position (car next-position)))))))
    (lambda ()
      (if (> percent plength)
	  (if (null? (cdr next-position))
	      ;; We need to change of pore
	      (let ((new-destination (select-next-pore next-pore *in*)))
		(if (null? new-destination) 
		    (begin 
		      (set! actual-pore (select-input-pore *in*))
		      (set! next-dbl (select-next-pore actual-pore *in*))
		      (set! next-pore (car next-dbl))
		      (set! pore-position (get-pore-position actual-pore *in*))
		      (set! next-position (list (cadr next-dbl) (get-pore-position next-pore *in*)))
		      (set! percent 0)
		      (set! plength (* scale-m (sqrt (apply + (map (lambda (p1 p2) (* (- p2 p1) (- p2 p1))) pore-position (car next-position)))))))
		    (begin
		      (set! actual-pore next-pore)
		      (set! next-dbl new-destination)
		      (set! next-pore (car next-dbl))
		      (set! pore-position (get-pore-position actual-pore *in*))
		      (set! next-position (list (cadr next-dbl) (get-pore-position next-pore *in*)))
		      (set! percent 0)
		      (set! plength (* scale-m (sqrt (apply + (map (lambda (p1 p2) (* (- p2 p1) (- p2 p1))) pore-position (car next-position)))))))))
	      ;; We go at the second half of the pore 
	      (begin		
		(set! pore-position (car next-position))
		(set! next-position (cdr next-position))
		(set! percent 0)
		(set! plength (* scale-m (sqrt (apply + (map (lambda (p1 p2) (* (- p2 p1) (- p2 p1))) pore-position (car next-position))))))
		)))
      (set! percent (+ percent 1))
      (list 
       (map (lambda (a b) (exact->inexact (+ a (* (- b a) (/ percent plength))))) pore-position (car next-position))
       (map (lambda (a b) (exact->inexact (+ a (* (- b a) (/ (+ percent 5) plength))))) pore-position (car next-position)))
      )))

(define go 
  ;; this function is called each time OpenDX run the loop.
  ;; On the first call, it read the input parameters
  (let ((p1 '())
	(p2 '())
	(l1 '())
	(l2 '())
	(pos1 '())
	(pos2 '())
	(dir1 '())
	(dir2 '()))
    (lambda ()
      (if (null? *in*) (get-dx-input ))
      (if (null? p1) 
	  (set! p1 (create-particule )))
      (if (null? p2) 
	  (set! p2 (create-particule )))
      (call-again ) ;; run me again ...
      (set! l1 (p1))
      (set! l2 (p2))
      (set! pos1 (car l1))
      (set! pos2 (car l2))
      (set! dir1 (cadr l1))
      (set! dir2 (cadr l2))
      ;; format *out* for output. 
      ;; The first output is a field with 3 array in it:
      ;; - position, containing the position of all balls;
      ;; - data, information to assign different colors;
      ;; - size, the size of each particule.
      ;; The second output is the position of the camera.
      ;; The third output is the direction of the camera.
      ;; The fourth output is the number of the turn. 
      (set! *out* (list (list (list "positions" (list pos1 pos2))
			      (list "data" (list (list 2) (list 1)) "dep" "positions")
			      (list "size" (list (list 1) (list 1)) "dep" "positions")
			      )
			(list pos1)
			(list dir1)
			turn))
      (set! turn (+ turn 1))
      (display *out*)
      (set-dx-output )
      )))
